/* File:      parm.C
   Date:      5/20/03 (originally written fall 2001 version)
   Author:    C. Paciorek
   Purpose:   code for parameter objects, including MCMC proposal, evaluation and updating
   Compile:   g++ -c -Wall -ansi
   Usage:     N/A
   Arguments: N/A
   Requires:  see just below
   Reference: none
*/

#include "base.h"
#include "parm.h"
#include "density.h"
#include "gasdev.h"
#include "ran1.h"
#include "cov_mat.h"
#include "dataset.h"
#include "transform.h"
#include "covariates.h"

///////////////////////////////////////////////////////
// Class nonspatial_parm
///////////////////////////////////////////////////////

nonspatial_parm::nonspatial_parm(int which_in,double omega_init,double omega_mean, double omega_sd, double prop_sd_in,transform_fxn_ptr transform_in){
  int accepted=ACCEPT;
  which=which_in;            // which parameter is it
  psi=new hyperparm_scal(omega_init,omega_mean,omega_sd,prop_sd_in,&normal_dens,transform_in);   // normal prior on transformed scale
  psi->evaluate_prior();     // initialize prior density
  psi->update(accepted);     // have value and value_stored be the same to start with
  num_accept=0;
  run_num_accept=0;
}

nonspatial_parm::~nonspatial_parm(void){
  delete psi;
}

int nonspatial_parm::accept(double logprior_diff){
  float unif=ran1(idum,GENERATE,"a");   // the "a" is just a dummy argument,not needed for GENERATE mode
  if(exp(data->evaluate_loglik_diff(which)+logprior_diff)>unif){
    return(ACCEPT);
  }
  else{
    return(REJECT);
  }
}


int nonspatial_parm::step(int fixed){
  int accepted=REJECT;
  double logprior_diff;
  if(!fixed){
    psi->propose(0.0);   // 0.0 indicates that the default prop_sd value should be used
    logprior_diff=psi->evaluate_prior();
    if(zero(logprior_diff-NEG_INFTY)){
      psi->update(REJECT);
    }
    else{
      accepted=accept(logprior_diff);
      num_accept+=accepted;
      run_num_accept+=accepted;
      psi->update(accepted);         // update value and value_stored
      data->update(accepted,which);  // update likelihood
    }
  }
  return(accepted);
}


////////////////////////////////////
// Class spatial_parm
////////////////////////////////////

spatial_parm::spatial_parm(int which_in,double* omega_init,double mu_init, double logsigma_init,double nu_init,double mu_mean, double mu_sd, double logsigma_mean, double logsigma_sd, double nu_mean, double nu_sd,double omega_prop_sd, double mu_prop_sd, double logsigma_prop_sd, double nu_propsd,transform_fxn_ptr transform_in){
  // this constructor is for the regression function and eigenprocesses; a separate spatial_parm with separate constructor (below) has the correlation structure for the eigenprocesses
  int accepted=ACCEPT;

  num_accept_omega=0;
  num_accept_mu=0;
  num_accept_sigma=0;
  num_accept_nu=0;
  run_num_accept_omega=0;
  run_num_accept_mu=0;
  run_num_accept_sigma=0;
  run_num_accept_nu=0;
  transform=transform_in;
  which=which_in;
  num_omegapoints=covariates->num_trainpoints;

  psi_full=new double[covariates->num_datapoints];

  // maximum value for eigenvalue process value at any location - this prevents poor mixing when eigenvalues get large and don't affect likelihood or prior much
  if(covariates->num_covars==1){
    lam_max=9.0;
  }
  else{  // have max depend on num_covars because for more covariates, it's harder to ignore a covariate unless eigenvalues get large
    lam_max=9.0*covariates->num_covars;
  }

  tmpvec=new double[num_omegapoints];
  tmpvec2=new double[num_omegapoints];

  tmpvec_test=new double[covariates->num_datapoints];
  tmpvec_test2=new double[covariates->num_datapoints];
    
  // vectors of gradients for Langevin calculations
  oldgrad=new double[num_omegapoints];
  newgrad=new double[num_omegapoints];

  oldgrad_exceeded=0;
  newgrad_exceeded=0;

  // vectors for PMC calculations
  LCsum=new double[num_omegapoints];
  Csum_diff=new double[num_omegapoints];

  // mean of process
  mu=new hyperparm_scal(mu_init,mu_mean,mu_sd,mu_prop_sd,&normal_dens,&identity);
  // log(sd) of process
  sigma=new hyperparm_scal(logsigma_init,logsigma_mean,logsigma_sd,logsigma_prop_sd,&normal_dens,&log_transform);

  // put init and propsd in here and took init out of omega
  ksi=new hyperparm_vec(num_omegapoints);  // process values on transformed scale
  omega=new hyperparm_vec(num_omegapoints,omega_init,omega_prop_sd); // a priori independent components on transformed scale
  psi=new hyperparm_vec(num_omegapoints);        // process values on original scale
  chi=new hyperparm_vec(num_omegapoints);        // decorrelated residuals about conditional posterior mean - used in PMC
  chinext=new hyperparm_vec(num_omegapoints);        // decorrelated residuals about proposed conditional posterior mean - used in PMC Hastings correction

  // smoothness parameter in Matern correlation for function
  if(which==FUN && prior==NON_STATIONARY){
    nu=new hyperparm_scal(nu_init,nu_mean,nu_sd,nu_propsd,&uniform_dens,&identity);
    nu->evaluate_prior();
    nu->update(accepted);
  }
  else{
    nu=0;
  }

  // these are not used in the function spatial_parm only in the kernel spatial_parm
  kappa=0;
  gamma=0;
  stat_cov=0;
  stat_cov_full=0;

  mu->evaluate_prior();
  sigma->evaluate_prior();
  omega->evaluate_prior();

  // set value_stored to be same as value
  mu->update(accepted);
  sigma->update(accepted);
  omega->update(accepted);
}

spatial_parm::spatial_parm(int which_in,  double* logkappa_init,double* gamma_init,double nu_init, double logkappa_parm1,double logkappa_parm2, double nu_mean, double nu_sd, double* logkappa_prop_sd, double* gamma_prop_sd,double nu_propsd){
  // this constructor is for the kernel structure - the prior covariance structure for the eigenprocesses 
  int p;
  int accepted=ACCEPT;

  num_accept_kappa=new int[covariates->num_covars];
  num_accept_gamma=new int[covariates->num_angles];
  num_accept_nu=0;
  run_num_accept_kappa=new int[covariates->num_covars];
  run_num_accept_gamma=new int[covariates->num_angles];
  run_num_accept_nu=0;

  for(p=0;p<covariates->num_covars;p++){
    num_accept_kappa[p]=0;
    run_num_accept_kappa[p]=0;
  }
  for(p=0;p<covariates->num_angles;p++){
    num_accept_gamma[p]=0;
    run_num_accept_gamma[p]=0;
  }

  // maximum value for eigenvalue process value at any location - this prevents poor mixing when eigenvalues get large and don't affect likelihood or prior much
  if(covariates->num_covars==1){
    lam_max=9.0;
  }
  else{  // have max depend on num_covars because for more covariates, it's harder to ignore a covariate unless eigenvalues get large
    lam_max=9.0*covariates->num_covars;
  }


  which=which_in;
  num_omegapoints=covariates->num_trainpoints;

  tmpvec=new double[num_omegapoints];
  tmpvec2=new double[num_omegapoints];

  // not needed since no process values in this construction
  tmpvec_test=0;
  tmpvec_test2=0;

  // vectors for post mean calculations
  LCsum=new double[num_omegapoints];
  Csum_diff=new double[num_omegapoints];

  // vectors for Langevin calculations  (not used for this construction)
  oldgrad=new double[num_omegapoints];
  newgrad=new double[num_omegapoints];

  oldgrad_exceeded=0;
  newgrad_exceeded=0;

  // not used in this construction of correlation structure for eigenprocesses
  mu=0; sigma=0; ksi=0; psi=0; omega=0; chi=0;

  nu=new hyperparm_scal(nu_init,nu_mean,nu_sd,nu_propsd,&uniform_dens,&identity);

  // scale parameters (sqrts of eigenvalues in prior structure for eigenprocesses
  kappa=new (hyperparm_scal*)[covariates->num_covars];
  kappa_dbl=new (double*)[covariates->num_covars];
  for(p=0;p<covariates->num_covars;p++){
    kappa[p]=new hyperparm_scal(logkappa_init[p],logkappa_parm1,logkappa_parm2,logkappa_prop_sd[p],&beta_dens,&log_transform);
    if(kappa[p]->base_value>logkappa_parm2 || kappa[p]->base_value<logkappa_parm1){  // make sure kappa is in parameter range 
      cerr << "Warning, kappa value not in range " << kappa[p]->value << " " << logkappa_parm1 << " " << logkappa_parm2 << "\n";
    }
    kappa_dbl[p]=&(kappa[p]->value);  // vector of values used in covariance matrix filling
  }
  for(p=0;p<covariates->num_covars;p++){
    kappa[p]->evaluate_prior();                  // initialize prior densities ...
    kappa[p]->update(accepted);                  // initialize stored values ...
  }    

  if(covariates->num_angles){
    // Givens angles in prior structure for eigenprocesses
    gamma=new (hyperparm_scal*)[covariates->num_angles];
    gamma_dbl=new (double*)[covariates->num_angles];  // vector of values used in covariance matrix filling
    gamma_max=new double[covariates->num_angles];
    gamma_min=new double[covariates->num_angles];

    // order allows code to cycle through powers of cosine prior density; different Givens angles have prior densities that are different powers of the cosine function so as to preserve rotational symmetry in the prior
    int order=0;
    int max=covariates->num_covars-1;
    for(p=0;p<covariates->num_angles;p++){
      if(order==max){
        order=0;
        max--;
      }
      gamma[p]=new hyperparm_scal(gamma_init[p],(double)order,0.0,gamma_prop_sd[p],&cos_dens,&identity);
      gamma_min[p]=-PI;
      gamma_max[p]=PI;
      order++;
      gamma_dbl[p]=&(gamma[p]->value);
    }
    for(p=0;p<covariates->num_angles;p++){
      gamma[p]->evaluate_prior();                  // initialize prior densities ...
      gamma[p]->update(accepted);                  // initialize stored values ...
    }    
  }
  else{
    gamma=0; gamma_dbl=0;
  }

  // shared stationary prior correlation for eigenprocesses

  // training set
  stat_cov=new sparse_cov_mat_stat(num_omegapoints,covariates->num_covars,kappa_dbl,gamma_dbl,&(nu->value));
  // training plus test set
  stat_cov_full=new sparse_cov_mat_stat(covariates->num_datapoints,covariates->num_covars,kappa_dbl,gamma_dbl,&(nu->value));


  nu->evaluate_prior();                  // initialize prior densities ...
  nu->update(accepted);                  // initialize stored values ...
}

spatial_parm::~spatial_parm(){
  if(mu){
    delete mu;
  }
  if(sigma){
    delete sigma;
  }
  if(omega){
    delete omega;
  }
  if(psi){
    delete psi;
  }
  if(ksi){
    delete ksi;
  }
  if(chi){
    delete chi;
  }
  if(kappa){
    for(int p=0;p<covariates->num_covars;p++){
      delete kappa[p];
    }
    delete [] kappa;
    delete [] kappa_dbl;
  }
  if(gamma){
    for(int p=0;p<covariates->num_angles;p++){
      delete gamma[p];
    }
    delete [] gamma;
    delete [] gamma_dbl;
  }
  if(which==KERNEL){
    delete stat_cov;
    delete stat_cov_full;
  }
  if(tmpvec_test){
    delete [] tmpvec_test;
  }
  if(tmpvec_test2){
    delete [] tmpvec_test2;
  }
  delete [] tmpvec;
  delete [] tmpvec2;
  delete [] LCsum;
  delete [] Csum_diff;
  delete [] oldgrad;
  delete [] newgrad;
}

// convert omega to ksi by multiplying by Cholesky, scaling and adding mean
void spatial_parm::omega_to_ksi(void){
  if(which==FUN && prior==NON_STATIONARY){
    nonstat_cov->Lmult(omega->value,tmpvec);
  }
  else{
    stat_cov->Lmult(omega->value,tmpvec);
  }
  for(int i=0;i<num_omegapoints;i++){
    ksi->value[i]=tmpvec[i]*sigma->value+mu->value;
  }
}

// transform ksi (on transformed scale) to psi (on original scale)
void spatial_parm::ksi_to_psi(void){
  for(int j=0;j<num_omegapoints;j++){
    psi->value[j]=transform(ksi->value[j]);
  }  
}

// consider acceptance when parameter not involved in likelihood
int spatial_parm::accept_nolik(double prior_diff){
  float unif=ran1(idum,GENERATE,"a"); 
  if(exp(prior_diff)>unif){ 
    return(ACCEPT);
  }
  else{
    return(REJECT);
  }
}

// consider acceptance when parameter is involved in likelihood
int spatial_parm::accept_lik(double logprior_diff){
  float unif=ran1(idum,GENERATE,"a"); 
  double loglik_diff=data->evaluate_loglik_diff(which);
  if(exp(loglik_diff+logprior_diff)>unif){ 
    return(ACCEPT);
  }
  else{
    return(REJECT);
  }
}


// add information unavailable at time of construction

void spatial_parm::add_child(sparse_cov_mat_stat* cov_in,sparse_cov_mat_stat* cov_full_in){
  // adds pointer to stationary covariance  to fun object
  stat_cov=cov_in;
  nonstat_cov=0;
  stat_cov_full=cov_full_in;
  nonstat_cov_full=0;

  // find ksi and psi, now that have covariance matrix and Cholesky
  omega_to_ksi();
  ksi_to_psi();
  ksi->update(ACCEPT);
  psi->update(ACCEPT);
  
  if(which==FUN_COVAR_EIGVAL){
    for(int j=0;j<num_omegapoints;j++){
      if(psi->value[j]>=lam_max) {
	//psi->value[j]=1.0;
	// just setting psi values doesn't change omega, so this makes no sense
	cerr << "Warning: some initial psi values for " << which << " are greater than 9*num_covars\n";
      }
    }
  }
  cout << setprecision(15) << "Starting values for " << which << ":" << endl;
  for(int j=0;j<num_omegapoints;j++){
    cout << psi->value[j] << " ";
  }
  cout << endl;
}

void spatial_parm::add_child(sparse_cov_mat_nonstat* cov_in,sparse_cov_mat_nonstat* cov_full_in){
  // adds pointer to nonstationary covariance to fun object
  nonstat_cov=cov_in;
  stat_cov=0;
  nonstat_cov_full=cov_full_in;
  stat_cov_full=0;

  // find ksi and psi, now that have covariance matrix and Cholesky
  omega_to_ksi();
  ksi_to_psi();
  ksi->update(ACCEPT);
  psi->update(ACCEPT);
  
  cout << "Starting values for " << which << ":" << endl;
  for(int j=0;j<num_omegapoints;j++){
    cout << psi->value[j] << " ";
  }
  cout << endl;
}

void spatial_parm::add_child(spatial_parm* fun_in){
  // adds pointer to fun to kernel object that controls eigenprocess correlation
    fun=fun_in;
}

void spatial_parm::add_child(spatial_parm** eigenval_in,spatial_parm** eigenvec_in){
  // adds pointers to eigenprocesses to kernel object
    eigenval=eigenval_in;
    eigenvec=eigenvec_in;
}

void spatial_parm::sample_full(void){
  // sample processes at test covariates
  int i;
  if(which==FUN){
    // find full covariance matrix
    if(prior==STATIONARY){
      stat_cov_full->fill(NOT_FIRST);
      stat_cov_full->gen_chol(NO_WARN);
    }
    else{
      nonstat_cov_full->fill(FUN_COVAR_EIGVEC,NOT_FIRST);
      nonstat_cov_full->gen_chol(NO_WARN);
    }
  }
  for(i=0;i<num_omegapoints;i++){
    tmpvec_test[i]=omega->value[i];
  }
  // generate N(0,1) deviates for prediction at test covariates
  for(i=num_omegapoints;i<covariates->num_datapoints;i++){
    tmpvec_test[i]=gasdev(idum);
  }


  // calculate process values on original scale

  if(which==FUN && prior==NON_STATIONARY){
    nonstat_cov_full->Lmult(tmpvec_test,tmpvec_test2);
  }
  else{
    stat_cov_full->Lmult(tmpvec_test,tmpvec_test2);
  }
  for(i=0;i<covariates->num_datapoints;i++){
    tmpvec_test[i]=tmpvec_test2[i]*sigma->value+mu->value;
  }
  for(i=0;i<covariates->num_datapoints;i++){
    psi_full[i]=transform(tmpvec_test[i]);
  }  
}
  

void spatial_parm::update(int accepted,int mode){
  if(which==FUN){
    if(prior==NON_STATIONARY){
      nonstat_cov->update(accepted);
    }
    else{
      stat_cov->update(accepted);
    }
    omega->update(accepted);
    ksi->update(accepted);
    psi->update(accepted);
    if(mode==POSTMEAN){  
      if(accepted){
        copy(covariates->num_trainpoints,chinext->value,chi->value);
      }
      chi->update(accepted);
    }
  }
  else{
    omega->update(accepted);
    ksi->update(accepted);
    psi->update(accepted);
  }
}

double spatial_parm::change(int caller,int mode){
  if(which==KERNEL){
    cerr << "Warning: change function only designed for use as method of regression function or eigenprocess objects\n";
  }

  double sd=omega->prop_sd*.0001; 
  // this is set to a small fraction of the omega prop sd so that the main factor affecting acceptance is the change in the hyperparameter and its direct effect on the process, not the additional noise added to omega

  if(which==FUN){
    // fill and get Cholesky for prior covariance and sum of prior covariance and data covariance
    if(prior==STATIONARY){
      stat_cov->fill(NOT_FIRST);
      stat_cov->gen_chol(NO_WARN);
    }
    else{
      nonstat_cov->fill(caller,NOT_FIRST);
      nonstat_cov->gen_chol(NO_WARN);
    }

    if(mode==POSTMEAN){
      // use PMC proposal
      ksi_propose(FALSE);
    }
    else{
      // use SIMPLE proposal
      omega->propose(sd);

      // prevents updating of elements whose columns are zeroed out
      if(prior==STATIONARY){
        stat_cov->Lcheck(omega->value,omega->value_stored);  
      }
      else{
        nonstat_cov->Lcheck(omega->value,omega->value_stored);  
      }

      omega_to_ksi();
    }

  }
  else{  // eigenprocess proposal
    omega->propose(sd);
    stat_cov->Lcheck(omega->value,omega->value_stored);  // prevents updating of elements whose columns are zeroed out
    omega_to_ksi();
  }
  double logprior_diff=omega->evaluate_prior();
  ksi_to_psi();
  return(logprior_diff);
}

double spatial_parm::ksi_propose(int update_chi){
  // does PMC proposal or updates chi vector 
  int i;
  double sd=omega->prop_sd*.0001; 
  // this is set to a small fraction of the omega prop sd so that the main factor affecting acceptance is the change in the hyperparameter and its direct effect on the process, not the additional noise added to omega

  int flag;

  // calculate fhat 
  data->ytoyprime();
  for(i=0;i<num_omegapoints;i++){
    tmpvec[i]=data->Yprime[i]-mu->value;
  }

  data->cov->fill();
  data->cov->gen_chol(WARN);
  flag=data->cov->Csolve(tmpvec,Csum_diff);
  if(prior==STATIONARY){
    stat_cov->Ltmult(Csum_diff,LCsum);
  }
  else{
    nonstat_cov->Ltmult(Csum_diff,LCsum);
  }
  if(flag){  // sum of prior covariance and data covariance is numerically singular, which should not happen since data covariance is diagonal
    cerr << "Parameters at non-invertibility: " << which << " " << sigma->value << " " << *(data->noise_sd) << endl;
  }

  if(update_chi){  // calculate chi based on omega
    for(i=0;i<num_omegapoints;i++){
      chi->value[i]=omega->value[i]-sigma->value*LCsum[i];
    }
    chi->update(ACCEPT);
  }
  else{
    // propose ksi using PMC proposal
    if(prior==STATIONARY){
      stat_cov->Rmult(Csum_diff,ksi->value);
      chi->propose(sd);
      stat_cov->Lcheck(chi->value,chi->value_stored);  // don't update zeroed out column elements
      stat_cov->Lmult(chi->value,tmpvec);      
    }
    else{
      nonstat_cov->Rmult(Csum_diff,ksi->value);
      chi->propose(sd);
      nonstat_cov->Lcheck(chi->value,chi->value_stored);  // don't update zeroed out column elements
      nonstat_cov->Lmult(chi->value,tmpvec);
    }
    for(i=0;i<num_omegapoints;i++){ // find proposed ksi
      ksi->value[i]=ksi->value[i]*sigma->value*sigma->value+mu->value+sigma->value*tmpvec[i];
    }
    // calculate proposed omega based on proposed chi
    for(i=0;i<num_omegapoints;i++){
      omega->value[i]=chi->value[i]+sigma->value*LCsum[i];
    }    


    // the remaining code in this function calculates values of chi (chinext) based on the proposed ksi values for use in Hastings ratio and updating

    ksi_to_psi();  // this makes calls to this function later on redundant, but I need it here to be able to calculate Y' and C_Y'
    data->ytoyprime();
    
    // create chistar0
    for(i=0;i<num_omegapoints;i++){
      tmpvec[i]=data->Yprime[i]-mu->value_stored;
    }
    
    data->cov->fill_old();
    data->cov->gen_chol(WARN);
    flag=data->cov->Csolve(tmpvec,Csum_diff);
    
    if(prior==STATIONARY){
      stat_cov->Ltmult_old(Csum_diff,LCsum);
    }
    else{
      nonstat_cov->Ltmult_old(Csum_diff,LCsum);
    }
    if(flag){
      cout << "Parameters at non-invertibility: " << which << " " << sigma->value_stored << " " << *(data->noise_sd) << endl;
    }
    for(i=0;i<num_omegapoints;i++){
      chinext->value_stored[i]=omega->value_stored[i]-sigma->value_stored*LCsum[i];
    }

    // create chistarstar
    for(i=0;i<num_omegapoints;i++){
      tmpvec[i]=data->Yprime[i]-mu->value;
    }
    
    data->cov->fill();
    data->cov->gen_chol(WARN);
    flag=data->cov->Csolve(tmpvec,Csum_diff);
    
    if(prior==STATIONARY){
      stat_cov->Ltmult(Csum_diff,LCsum);
    }
    else{
      nonstat_cov->Ltmult(Csum_diff,LCsum);
    }
    if(flag){
      cout << "Parameters at non-invertibility: " << which << " " << sigma->value << " " << *(data->noise_sd) << endl;
    }
    for(i=0;i<num_omegapoints;i++){
      chinext->value[i]=omega->value[i]-sigma->value*LCsum[i];
    }
    // calculate Hastings ratio
    for(i=0;i<num_omegapoints;i++){
      logprior_diff+=(-0.5)*pow(chinext->value_stored[i]-chinext->value[i],2.0)+0.5*pow(chi->value_stored[i]-chi->value[i],2.0);
    }
  }
  return(logprior_diff);
}

void spatial_parm::step(int mode,int langevin,int kappa_fixed, int gamma_fixed,int mu_fixed, int sigma_fixed,int nu_fixed,int omega_fixed){
  // do a single round of the MCMC proposals; arguments indicate style of proposal (SIMPLE/POSTMEAN,LANGEVIN/NON_LANGEVIN) and whether the hyperparameters are fixed or not
  int accepted=0;
  int p,i,q;
  double sd;
  if(which!=KERNEL){
    sd=omega->prop_sd*0.0001;  // omega doesn't exist if which==KERNEL so can't access prop_sd
  }
  if(langevin && which!=FUN){
    cerr << "Langevin proposal implemented only for regression function vector\n";
  }
  else{
    if(!nu_fixed && nu){

      // change proposal sd depending on current value of nu
      double tmp_sd,tmp_sdstar;
      tmp_sd=nu->prop_sd;
      if(nu->value<nu_lowval){
          tmp_sd*=lownu_scale;
        }
      if(nu->value>nu_highval){
          tmp_sd*=highnu_scale;
      }

      nu->propose(tmp_sd);
      logprior_diff=nu->evaluate_prior();
      if(zero(logprior_diff-NEG_INFTY)){
        nu->update(REJECT);
      }
      else{
        // adjust acceptance ratio for Hastings style proposal
        tmp_sdstar=nu->prop_sd;
        if(nu->value<nu_lowval){
          tmp_sdstar*=lownu_scale;
        }
        if(nu->value>nu_highval){
          tmp_sdstar*=highnu_scale;
        }
        logprior_diff+=log(tmp_sd)+0.5*pow(nu->base_value-nu->base_value_stored,2.0)/(tmp_sd*tmp_sd)-log(tmp_sdstar)-0.5*pow(nu->base_value-nu->base_value_stored,2.0)/(tmp_sdstar*tmp_sdstar);

        // propose function conditional on hyperparameter proposal
        if(which==FUN && prior==NON_STATIONARY){
          logprior_diff+=change(which,mode);
        }
        // propose function conditional on correlation structure proposal
        if(which==KERNEL && prior==STATIONARY){
          logprior_diff+=fun->change(which,mode);
        }
        // propose eigenprocesses conditional on shared correlation strcuture and propose function conditional on the proposed eigenstructure
        if(which==KERNEL && prior==NON_STATIONARY){
          stat_cov->fill(NOT_FIRST);
          stat_cov->gen_chol(NO_WARN);
          for(p=0;p<covariates->num_covars;p++){
            logprior_diff+=eigenval[p]->change(which,mode);
          }
          for(p=0;p<covariates->num_eigenvecs;p++){
            logprior_diff+=eigenvec[p]->change(which,mode);
          }
          logprior_diff+=fun->change(which,mode);
        }
        accepted=accept_lik(logprior_diff);

        // don't let eigenvalues get too big
        if(which==KERNEL && prior==NON_STATIONARY){
          for(p=0;p<covariates->num_covars;p++){
            for(int i=0;i<num_omegapoints;i++){
              if(eigenval[p]->psi->value[i]>lam_max){
                accepted=REJECT;
              }
            }
          }
        }
        if(which==FUN_COVAR_EIGVAL){
          for(int i=0;i<num_omegapoints;i++){
            if(psi->value[i]>lam_max){
              accepted=REJECT;
            }
          }
        }

        // updating
        nu->update(accepted);
        data->update(accepted,which);
        if(which==KERNEL){
          fun->update(accepted,mode);
        }
        else{
          update(accepted,mode);
        }
        if(which==KERNEL && prior==NON_STATIONARY){
          stat_cov->update(accepted);
          for(p=0;p<covariates->num_covars;p++){
            eigenval[p]->update(accepted,mode);
          }
          for(p=0;p<covariates->num_eigenvecs;p++){
            eigenvec[p]->update(accepted,mode);
          }
        }
        num_accept_nu+=accepted;
        run_num_accept_nu+=accepted;
      }
    }
    if(!gamma_fixed && gamma){  // Givens angle proposals
      // propose each angle in turn
      for(p=0;p<covariates->num_angles;p++){
        accepted=ACCEPT;
        gamma[p]->propose(0.0);

        // keep Givens angles within parameter range
        while(gamma[p]->base_value<gamma_min[p]){
          gamma[p]->base_value+=2.0*PI;
          gamma[p]->value=gamma[p]->base_value;
        }
        while(gamma[p]->base_value>gamma_max[p]){
          gamma[p]->base_value-=2.0*PI;
          gamma[p]->value=gamma[p]->base_value;
        }

        logprior_diff=gamma[p]->evaluate_prior();
        if(zero(logprior_diff-NEG_INFTY)){  // reject if value outside support
          accepted=REJECT;
        }
        if(!accepted){
            gamma[p]->update(accepted);
        }
        else{
          // propose eigenprocesses conditional on shared correlation structure and function conditional on eigenprocesses
          if(prior==NON_STATIONARY){
            stat_cov->fill(NOT_FIRST);
            stat_cov->gen_chol(NO_WARN);
            for(q=0;q<covariates->num_covars;q++){
              logprior_diff+=eigenval[q]->change(which,mode);
            }
            for(q=0;q<covariates->num_eigenvecs;q++){
              logprior_diff+=eigenvec[q]->change(which,mode);
            }

          }
          logprior_diff+=fun->change(which,mode);

          accepted=accept_lik(logprior_diff);
        
          // don't let eigenvalues get too big
          if(which==KERNEL && prior==NON_STATIONARY){
            for(q=0;q<covariates->num_covars;q++){
              for(int i=0;i<num_omegapoints;i++){
                if(eigenval[q]->psi->value[i]>lam_max){
                  accepted=REJECT;
                }
              }
            }
          }
          if(which==FUN_COVAR_EIGVAL){
            for(int i=0;i<num_omegapoints;i++){
              if(psi->value[i]>lam_max){
                accepted=REJECT;
              }
            }
          }

          // updating
          gamma[p]->update(accepted);
          fun->update(accepted,mode);
          data->update(accepted,which);

          if(prior==NON_STATIONARY){
            stat_cov->update(accepted);
            for(q=0;q<covariates->num_covars;q++){
              eigenval[q]->update(accepted,mode);
            }
            for(q=0;q<covariates->num_eigenvecs;q++){
              eigenvec[q]->update(accepted,mode);
            }
          }

          num_accept_gamma[p]+=accepted;
          run_num_accept_gamma[p]+=accepted;
        }
      }
    }
    if(!kappa_fixed && kappa){  // correlation scale parameters for stationary function or for eigenprocesses
      double tmp_sd,tmp_sdstar;
      // propose each parameter in turn
      for(p=0;p<covariates->num_covars;p++){
        accepted=ACCEPT;

        // adjust proposal sd depending on current value
        tmp_sd=kappa[p]->prop_sd;
        
        if(kappa[p]->value>kappa_highval){
          tmp_sd*=highkappa_scale;
        }
        if(kappa[p]->value<kappa_lowval){
          tmp_sd*=lowkappa_scale;
        }
        
        kappa[p]->propose(tmp_sd);

        logprior_diff=kappa[p]->evaluate_prior();
        if(zero(logprior_diff-NEG_INFTY)){  // reject if value outside support
          accepted=REJECT;
        }
        if(!accepted){
            kappa[p]->update(accepted);
        }
        else{
          // do a Hastings adjustment to the acceptance ratio
          tmp_sdstar=kappa[p]->prop_sd;
          if(kappa[p]->value>kappa_highval){
            tmp_sdstar*=highkappa_scale;
          }
          if(kappa[p]->value<kappa_lowval){
            tmp_sdstar*=lowkappa_scale;
          }
          logprior_diff+=log(tmp_sd)+0.5*pow(kappa[p]->base_value-kappa[p]->base_value_stored,2.0)/(tmp_sd*tmp_sd)-log(tmp_sdstar)-0.5*pow(kappa[p]->base_value-kappa[p]->base_value_stored,2.0)/(tmp_sdstar*tmp_sdstar);

          // propose eigenprocesses conditional on proposed correlation structure and function conditional on eigenprocesses
          if(prior==NON_STATIONARY){
            stat_cov->fill(NOT_FIRST);
            stat_cov->gen_chol(NO_WARN);
            for(q=0;q<covariates->num_covars;q++){
              logprior_diff+=eigenval[q]->change(which,mode);
            }
            for(q=0;q<covariates->num_eigenvecs;q++){
              logprior_diff+=eigenvec[q]->change(which,mode);
            }

          }
          logprior_diff+=fun->change(which,mode);

          accepted=accept_lik(logprior_diff);
          
          // don't let eigenvalues get too big
          if(which==KERNEL && prior==NON_STATIONARY){
            for(q=0;q<covariates->num_covars;q++){
              for(int i=0;i<num_omegapoints;i++){
                if(eigenval[q]->psi->value[i]>lam_max){
                  accepted=REJECT;
                }
              }
            }
          }
          if(which==FUN_COVAR_EIGVAL){
            for(int i=0;i<num_omegapoints;i++){
              if(psi->value[i]>lam_max){
                accepted=REJECT;
              }
            }
          }
          
          // updating
          kappa[p]->update(accepted);
          fun->update(accepted,mode);
          data->update(accepted,which);

          if(prior==NON_STATIONARY){
            stat_cov->update(accepted);
            for(q=0;q<covariates->num_covars;q++){
              eigenval[q]->update(accepted,mode);
            }
            for(q=0;q<covariates->num_eigenvecs;q++){
              eigenvec[q]->update(accepted,mode);
            }
          }

          num_accept_kappa[p]+=accepted;
          run_num_accept_kappa[p]+=accepted;
        }
      }
    }
    if(!mu_fixed && mu){  // mean hyperparameter
      mu->propose(0.0);
      logprior_diff=mu->evaluate_prior();

      if(which==FUN){      // don't use change() function because that does a lot of covariance calcs that aren't needed when proposing only mu
        if(mode==POSTMEAN){
          logprior_diff+=ksi_propose(FALSE);  // PMC proposal
        }
        else{  // SIMPLE proposal (i.e., centered-joint)
          omega->propose(sd);
          if(prior==STATIONARY){
            stat_cov->Lcheck(omega->value,omega->value_stored);
          }
          else{
            nonstat_cov->Lcheck(omega->value,omega->value_stored);
          }
          omega_to_ksi();  
        }
        logprior_diff+=omega->evaluate_prior();
        ksi_to_psi();
      }
      else{  // eigenprocesses (actually only eigenvalue processes since eigenvector processes have fixed mu), use SIMPLE proposal
        omega->propose(sd);
        stat_cov->Lcheck(omega->value,omega->value_stored);
        omega_to_ksi();  
        logprior_diff+=omega->evaluate_prior();
        ksi_to_psi();
        // change function conditional on eigenprocesses
        logprior_diff+=fun->change(which,mode);
      }

      accepted=accept_lik(logprior_diff);
      
      // don't let eigenvalues get too big
      if(which==KERNEL && prior==NON_STATIONARY){
        for(q=0;q<covariates->num_covars;q++){
          for(int i=0;i<num_omegapoints;i++){
            if(eigenval[q]->psi->value[i]>lam_max){
              accepted=REJECT;
            }
          }
        }
      }
      if(which==FUN_COVAR_EIGVAL){
        for(int i=0;i<num_omegapoints;i++){
          if(psi->value[i]>lam_max){
            accepted=REJECT;
          }
        }
      }
      
      // updating
      mu->update(accepted);
      omega->update(accepted);
      ksi->update(accepted);
      psi->update(accepted);
      if(which==FUN && mode==POSTMEAN){
        if(accepted){
          copy(covariates->num_trainpoints,chinext->value,chi->value);
        }
        chi->update(accepted);
      }
      if(which!=FUN){
        fun->update(accepted,mode);
      }
      data->update(accepted,which);
      num_accept_mu+=accepted;
      run_num_accept_mu+=accepted;
    }
    if(!sigma_fixed){
      sigma->propose(0.0);
      logprior_diff=sigma->evaluate_prior();

      if(which==FUN){  // this does some unnecessary covariance  calculations since changing sigma does not change correlation matrix, but I have left them in for code readability
        logprior_diff+=change(which,mode);  // PMC proposal
      }
      else{  // eigenprocesses -- but code is not actually used because sigma is fixed in eigenprocesses; SIMPLE proposal
        omega->propose(sd);
        if(prior==STATIONARY){
          stat_cov->Lcheck(omega->value,omega->value_stored);
        }
        else{
          nonstat_cov->Lcheck(omega->value,omega->value_stored); 
          }
        omega_to_ksi();
        logprior_diff+=omega->evaluate_prior();
        ksi_to_psi();
        // change function conditional on eigenprocesses
        logprior_diff+=fun->change(which,mode);
      }
      accepted=accept_lik(logprior_diff);
      
      // don't let eigenvalues get too big
      if(which==KERNEL && prior==NON_STATIONARY){
        for(q=0;q<covariates->num_covars;q++){
          for(int i=0;i<num_omegapoints;i++){
            if(eigenval[q]->psi->value[i]>lam_max){
              accepted=REJECT;
            }
          }
        }
      }
      if(which==FUN_COVAR_EIGVAL){
        for(int i=0;i<num_omegapoints;i++){
          if(psi->value[i]>lam_max){
            accepted=REJECT;
          }
        }
      }
      
      // updating
      sigma->update(accepted);
      omega->update(accepted);
      ksi->update(accepted);
      psi->update(accepted);
      if(which==FUN && mode==POSTMEAN){
        if(accepted){
          copy(covariates->num_trainpoints,chinext->value,chi->value);
        }
        chi->update(accepted);
      }
      if(which!=FUN){
        fun->update(accepted,mode);
      }
      data->update(accepted,which);
      num_accept_sigma+=accepted;
      run_num_accept_sigma+=accepted;
    }

    if(!omega_fixed){  // propose process on own without changing hyperparameters
      if(which==FUN){
        if(!langevin){
          omega->propose(0.0);
          if(prior==STATIONARY){
            stat_cov->Lcheck(omega->value,omega->value_stored);
          }
          else{
            nonstat_cov->Lcheck(omega->value,omega->value_stored);
          }
          logprior_diff=omega->evaluate_prior();
          
          omega_to_ksi();
          ksi_to_psi();
          accepted=accept_lik(logprior_diff);
          
          if(mode==POSTMEAN && accepted){
            // need to update chi based on changing Y' and C_Y'
            ksi_propose(TRUE);
          }
          
          omega->update(accepted);
          ksi->update(accepted);
          psi->update(accepted);
          data->update(accepted,which);  // update likelihood and data covariance matrix  -- but only for function
          num_accept_omega+=accepted;
          run_num_accept_omega+=accepted;
        }
        else{  // Langevin proposal for omega/psi based on gradient for omega (gradient for psi involves inverse of Cholesky, so can't do it that way)
          
          // truncate gradient when gets too big
          double H=50.0;
          
          // current gradient
          for(i=0;i<num_omegapoints;i++){
            tmpvec[i]=data->Y[i]-psi->value[i];
          }
          if(prior==STATIONARY){
            stat_cov->Ltmult(tmpvec,oldgrad);
          }
          else{
            nonstat_cov->Ltmult(tmpvec,oldgrad);
          }
          int flag=0;
          for(i=0;i<num_omegapoints;i++){
            oldgrad[i]=sigma->value*oldgrad[i]/(*(data->noise_sd)*(*(data->noise_sd)))-omega->value[i];
            // truncate gradient
            if(oldgrad[i]>H){
              flag=1;
              oldgrad[i]=H;
            }
            if(oldgrad[i]<(-H)){
              oldgrad[i]=-H;
              flag=1;
            }
          }
          if(flag){
            oldgrad_exceeded++;
          }
          
          omega->propose(0.0);
          if(prior==STATIONARY){
            stat_cov->Lcheck(omega->value,omega->value_stored);
          }
          else{
            nonstat_cov->Lcheck(omega->value,omega->value_stored);
          }
          for(i=0;i<num_omegapoints;i++){
            omega->value[i]+=(omega->prop_sd*omega->prop_sd)*oldgrad[i]/2.0;
          }
          
          omega_to_ksi();
          ksi_to_psi();
          
          // gradient at proposed psi/omega
          for(i=0;i<num_omegapoints;i++){
            tmpvec[i]=data->Y[i]-psi->value[i];
          }
          if(prior==STATIONARY){
            stat_cov->Ltmult(tmpvec,newgrad);
          }
          else{
            nonstat_cov->Ltmult(tmpvec,newgrad);
          }
          flag=0;
          for(i=0;i<num_omegapoints;i++){
            newgrad[i]=sigma->value*newgrad[i]/(*(data->noise_sd)*(*(data->noise_sd)))-omega->value[i];
            // truncate gradient
            if(newgrad[i]>H){
              newgrad[i]=H;
              flag=1;
            }
            if(newgrad[i]<(-H)){
              newgrad[i]=-H;
              flag=1;
            }
          }
          
          if(flag){
            newgrad_exceeded++;
          }
          
          logprior_diff=omega->evaluate_prior();
          
          // make Hastings adjustments to acceptance ratio
          for(i=0;i<num_omegapoints;i++){
            tmpvec[i]=omega->value_stored[i]-omega->value[i]-(omega->prop_sd*omega->prop_sd/2.0)*newgrad[i];
            tmpvec2[i]=omega->value[i]-omega->value_stored[i]-(omega->prop_sd*omega->prop_sd/2.0)*oldgrad[i];
            logprior_diff+=-0.5*(tmpvec[i]*tmpvec[i]-tmpvec2[i]*tmpvec2[i])/(omega->prop_sd*omega->prop_sd);
          }
          
          accepted=accept_lik(logprior_diff);
          
          if(mode==POSTMEAN && accepted){
            // need to update chi based on changing Y' and C_Y'
            ksi_propose(TRUE);
          }

          omega->update(accepted);
          ksi->update(accepted);
          psi->update(accepted);
          
          data->update(accepted,which);  // update likelihood and data covariance matrix  -- but only for function
          num_accept_omega+=accepted;
          run_num_accept_omega+=accepted;
        }
      }
      else{  // propose eigenprocess on own without hyperparameters; change function conditional on eigenprocess
        omega->propose(0.0);
        stat_cov->Lcheck(omega->value,omega->value_stored);  // prevents updating of elements whose columns are zeroed out
        logprior_diff=omega->evaluate_prior();   
        omega_to_ksi();
        ksi_to_psi();
        
        // change function conditional on eigenprocess
        logprior_diff+=fun->change(which,mode);
        
        accepted=accept_lik(logprior_diff);
        
        // don't let eigenvalues get too big
        if(which==FUN_COVAR_EIGVAL){
          for(int i=0;i<num_omegapoints;i++){
            if(psi->value[i]>lam_max){
              accepted=REJECT;
            }
          }
        }

        // updating
        omega->update(accepted);
        ksi->update(accepted);
        psi->update(accepted);
        
        fun->update(accepted,mode);
        data->update(accepted,which);
        
        num_accept_omega+=accepted;
        run_num_accept_omega+=accepted;
      }
    }
  }
}

////////////////////////////////////////
// Class hyperparm_vec  
////////////////////////////////////////

hyperparm_vec::hyperparm_vec(int size_in,double* init,double prop_sd_in){
  // initialization for omega vector
  prop_sd=prop_sd_in;   
  size=size_in;
  value=init;             // value and init share storage
  value_stored=new double[size];
}

hyperparm_vec::hyperparm_vec(int size_in){
  // initialization for parameter not directly initialized or sampled during MCMC (ksi, psi, chi)
  size=size_in;                                    
  value=new double[size];
  value_stored=new double[size];
}

hyperparm_vec::~hyperparm_vec()  {
  delete [] value;         // init is deleted here, NOT where it was allocated in chain.C
  delete [] value_stored;
}

void hyperparm_vec::propose(double alternate_prop_sd){
  double tmp=prop_sd;
  if(!zero(alternate_prop_sd)){
    tmp=alternate_prop_sd;
  }
  for(int i=0;i<size;i++){
    value[i]+=tmp*gasdev(idum);  // normal proposal 
  }
}

double hyperparm_vec::evaluate_prior(){          // MVN(0,1) prior
    logprior=mvn_dens(size,value,1.0);
    if(zero(logprior-DENSITY_ERROR)) { cerr << "Warning: Possible error in log density for vector: " << logprior << " " << value[0] << " " << value[1]  << endl;} 
  return(logprior-logprior_stored);
}

void hyperparm_vec::update(int accepted)       
{
  if(accepted){                                // store newly-accepted values
    logprior_stored=logprior;
    copy(size,value,value_stored);
  }
  else{                                        // revert to stored values
    logprior=logprior_stored;
    copy(size,value_stored,value);
  }
}


////////////////////////////////////
// Class hyperparm_scal
////////////////////////////////////


hyperparm_scal::hyperparm_scal(double init,double parm1_in,double parm2_in,double prop_sd_in,prior_dens_fxn_ptr prior_dens_fxn_in,transform_fxn_ptr transform_in){
  transform=transform_in;
  prior_dens_fxn=prior_dens_fxn_in;
  prop_sd=prop_sd_in;
  if(zero(prop_sd_in)){  // don't sample parameters with proposal sd of zero
    fixed=TRUE;
  }
  else{
    fixed=FALSE;
  }
  base_value=init;             // base_value is on transformed scale
  value=transform(base_value);
  parm1=parm1_in;
  parm2=parm2_in;
}

void hyperparm_scal::propose(double alternate_prop_sd){      // normal proposal density on transformed scale                 
  if(!zero(alternate_prop_sd)){
    base_value+=alternate_prop_sd*gasdev(idum);
  }
  else{
    base_value+=prop_sd*gasdev(idum);   
  }
  value=transform(base_value);
}

void hyperparm_scal::update(int accepted){
  if(accepted){                                      // store newly-accepted values
    logprior_stored=logprior;
    value_stored=value;
    base_value_stored=base_value;                    // irrelevant if no transform 
  }
  else{                                              // revert to stored values
    logprior=logprior_stored;
    value=value_stored;
    base_value=base_value_stored;
  }
}

double hyperparm_scal::evaluate_prior(){
  logprior=prior_dens_fxn(base_value,parm1,parm2);           // normal prior density (for transformed RV)
  if(zero(logprior-DENSITY_ERROR)) { 
    cerr << "Warning: Possible error in log density for scalar: " << value << " " << parm1 << " " << parm2 << " " << base_value << " " << logprior << endl;
  } 
  if(zero(logprior-NEG_INFTY)){  // code in spatial_parm, nonspatial_parm know to check for NEG_INFTY
    return(logprior);
  }
  else{
    return(logprior-logprior_stored);
  }
}









